"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Solver = void 0;
const goal_queue_1 = require("../goal-queue");
const subst_1 = require("../subst");
// NOTE In implementation, we use array of queues to do search,
//   but we should be thinking in terms of tree instead of queues,
//   only by doing so, we can have a clear understanding of the implementation.
class Solver {
    constructor(input, opts) {
        this.count = 0;
        this.queues = input.queues;
        this.limit = opts.limit;
    }
    static forGoals(goals, opts) {
        const queues = [new goal_queue_1.GoalQueue(subst_1.Subst.create(), goals)];
        const searching = new Solver({ queues }, opts);
        return searching;
    }
    next() {
        while (true) {
            if (this.limit !== undefined && this.count === this.limit)
                return null;
            const queue = this.queues.shift();
            if (queue === undefined)
                return null;
            const queues = queue.step();
            if (queues === null) {
                this.count++;
                return queue.subst;
            }
            // NOTE about searching
            // push front |   depth first
            // push back  | breadth first
            this.queues.push(...queues);
        }
    }
    find() {
        const solutions = [];
        while (true) {
            const subst = this.next();
            if (subst === null)
                break;
            solutions.push(subst);
        }
        return solutions;
    }
}
exports.Solver = Solver;
