"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Subst = void 0;
const value_1 = require("../value");
const ut = __importStar(require("../ut"));
class Subst {
    constructor(map) {
        this.map = map;
    }
    static create() {
        const map = new Map();
        return new Subst(map);
    }
    clone() {
        const map = new Map(this.map);
        return new Subst(map);
    }
    extend(v, value) {
        const subst = this.clone();
        subst.map.set(v, value);
        return subst;
    }
    find(v) {
        return this.map.get(v);
    }
    walk(value) {
        while (value instanceof value_1.Var) {
            const found = this.find(value);
            if (found === undefined) {
                return value;
            }
            else {
                value = found;
            }
        }
        return value;
    }
    occur(v, value) {
        value = this.walk(value);
        if (value instanceof value_1.Var) {
            return value === v;
        }
        else if (value instanceof Array) {
            for (let e of value) {
                if (this.occur(v, e)) {
                    return true;
                }
            }
            return false;
        }
        else if (value_1.isObject(value)) {
            for (let k in value) {
                if (this.occur(v, value[k])) {
                    return true;
                }
            }
            return false;
        }
        else {
            return false;
        }
    }
    unify(x, y) {
        x = this.walk(x);
        y = this.walk(y);
        if (x instanceof value_1.Var && y instanceof value_1.Var && x === y) {
            return this;
        }
        else if (x instanceof value_1.Var) {
            if (this.occur(x, y)) {
                return null;
            }
            else {
                return this.extend(x, y);
            }
        }
        else if (y instanceof value_1.Var) {
            if (this.occur(y, x)) {
                return null;
            }
            else {
                return this.extend(y, x);
            }
        }
        else if (x instanceof Array && y instanceof Array) {
            return this.unifyArray(x, y);
        }
        else if (value_1.isObject(x) && value_1.isObject(y)) {
            return this.unifyObject(x, y);
        }
        else if (x === y) {
            return this;
        }
        else {
            return null;
        }
    }
    unifyObject(xs, ys) {
        if (Object.keys(xs).length >= Object.keys(ys).length) {
            return this.coverObject(xs, ys);
        }
        else {
            return this.coverObject(ys, xs);
        }
    }
    coverObject(xs, ys) {
        let subst = this;
        for (let k in ys) {
            if (xs[k] === undefined) {
                return null;
            }
            subst = subst.unify(xs[k], ys[k]);
            if (subst === null) {
                return null;
            }
        }
        return subst;
    }
    unifyArray(xs, ys) {
        let subst = this;
        if (xs.length !== ys.length) {
            return null;
        }
        let i = 0;
        while (i < xs.length) {
            subst = subst.unify(xs[i], ys[i]);
            if (subst === null) {
                return null;
            }
            else {
                i++;
            }
        }
        return subst;
    }
    deepWalk(x) {
        x = this.walk(x);
        if (x instanceof value_1.Var) {
            return x;
        }
        else if (x instanceof Array) {
            return x.map((e) => this.deepWalk(e));
        }
        else if (value_1.isObject(x)) {
            const y = {};
            for (let k in x) {
                y[k] = this.deepWalk(x[k]);
            }
            return y;
        }
        else {
            return x;
        }
    }
    reify(x) {
        return this.deepWalk(x);
    }
    // NOTE for testing
    assertFound(v, value) {
        const found = this.find(v);
        if (found === undefined) {
            throw new Error([`I can not find var: ${JSON.stringify(v)}`].join("\n"));
        }
        if (!ut.equal(found, value)) {
            throw new Error([
                `I expect to found value: ${JSON.stringify(value)}`,
                `  var: ${JSON.stringify(v)}`,
                `  found: ${JSON.stringify(found)}`,
            ].join("\n"));
        }
    }
}
exports.Subst = Subst;
