"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Table = void 0;
const value_1 = require("../value");
const solver_1 = require("../solver");
const Goals = __importStar(require("../goals"));
const Clauses = __importStar(require("../clauses"));
const ut = __importStar(require("../ut"));
const ty_1 = __importStar(require("@xieyuheng/ty"));
// NOTE Our table is like prolog's predicate.
// - We define predicate by writing down Horn clauses.
// - A predicate describe relation between its arguments.
// - Predicate can be defined by function and vice versa,
//   however predicate is good at expressing relations,
//   because it make no distinction between input and output.
// - A clause is either a rule or a fact.
//   The clauses that constitute a predicate denote logical alternatives:
//   If any clause is true, then the whole predicate is true.
class Table {
    constructor(opts) {
        this.clauses = [];
        this.name = opts.name;
        this.schema = opts.schema;
    }
    i(data, premises) {
        if (premises) {
            this.clauses.push(Clauses.Rule.create({ data, premises }));
        }
        else {
            this.clauses.push(Clauses.Fact.create({ data }));
        }
    }
    o(data) {
        return new Goals.Relation({ table: this, data });
    }
    query(data, opts = {}) {
        data = value_1.freshenValue(data);
        const solver = solver_1.Solver.forGoals([this.o(data)]);
        const solutions = solver.solve({ limit: opts.limit });
        const results = solutions.map((subst) => subst.reify(data));
        return results;
    }
    find(data, varSchemas, opts = {}) {
        data = value_1.freshenValue(data);
        const solver = solver_1.Solver.forGoals([this.o(data)]);
        const solutions = solver.solve({ limit: opts.limit });
        const results = [];
        const v = value_1.Var.createVarFinder(value_1.Var.extractVarMap(data));
        const varEntries = Object.keys(varSchemas).map((name) => [
            name,
            v([name]),
        ]);
        for (const subst of solutions) {
            const result = subst.reify(Object.fromEntries(varEntries));
            try {
                results.push(ty_1.default.object(varSchemas).validate(result));
            }
            catch (error) {
                if (error instanceof ty_1.Errors.InvalidData && error.data instanceof value_1.Var) {
                    // NOTE Ok
                }
                else {
                    throw error;
                }
            }
        }
        return results;
    }
    get(data) {
        const [result] = this.query(data, { limit: 1 });
        return result;
    }
    assert(data) {
        const result = this.get(data);
        if (result === undefined) {
            throw new Error([
                `An assertion is made that the data is in the table,`,
                `but it is actually not in the table.`,
                `  table: ${this.name}`,
                `  data: ${JSON.stringify(data)}`,
            ].join("\n"));
        }
    }
    assertNot(data) {
        const result = this.get(data);
        if (result !== undefined) {
            throw new Error([
                `An assertion is made that the data is **not** in the table,`,
                `but it is actually in the table.`,
                `  table: ${this.name}`,
                `  data: ${JSON.stringify(data)}`,
            ].join("\n"));
        }
    }
    assertQueryResults(data, results) {
        const found = this.query(data);
        if (!ut.equal(found, results)) {
            throw new Error([
                `An assertion is made on the query results,`,
                `but the found results is not equal to the asserted results.`,
                `  table: ${this.name}`,
                `  data: ${JSON.stringify(data)}`,
                `  found    results: ${JSON.stringify(found)}`,
                `  asserted results: ${JSON.stringify(results)}`,
            ].join("\n"));
        }
    }
}
exports.Table = Table;
