"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.freshenValue = exports.isObject = void 0;
const value_1 = require("../value");
function isObject(x) {
    return typeof x === "object" && x !== null && !(x instanceof Array);
}
exports.isObject = isObject;
// NOTE side-effect on vars
function freshenValue(value, vars = new Map()) {
    if (value instanceof value_1.Var) {
        // NOTE We should use name as the key,
        //   because in the same scope,
        //   variables of the same name are the same variable.
        const found = vars.get(value.name);
        if (found !== undefined) {
            return found;
        }
        else {
            const v = new value_1.Var(value.name);
            vars.set(v.name, v);
            return v;
        }
    }
    else if (value instanceof Array) {
        return value.map((e) => freshenValue(e, vars));
    }
    else if (isObject(value)) {
        let obj = {};
        for (const k in value) {
            obj[k] = freshenValue(value[k], vars);
        }
        return obj;
    }
    else {
        return value;
    }
}
exports.freshenValue = freshenValue;
